
/* 	Persona 4: DAN & Persona 5 model and animation importer (*.gmd; *.gfs; *.gap) written in MaxScript
	Version 0.8.1
	Script by TGE. Please give credit where is due.
	Special thanks to PolarSoda, mrdudesworkshop
*/

FileIn "MaxScriptExtensions\\Utils.ms"
FileIn "MaxScriptExtensions\\FileUtils.ms"
FileIn "MaxScriptExtensions\\FileStreamReader.ms"
FileIn "MaxScriptExtensions\\DDS.ms"
FileIn "MaxScriptExtensions\\Stack.ms"
FileIn "GFSFormatDeclaration.ms"

-- Loader for GFS formatted files
struct GFSFileLoader
(
	private m_Gfs				= undefined,
	private m_FileInfo			= undefined,
	private m_Reader			= undefined,
	private m_LoadExternalTex	= false,
	
	public ApplyNormals 				= true,
	public ApplyColors			 	 	= true,
	public ApplySkin 				 	= true,
	public ApplyYToZUp	 	 			= true,
	public OverwriteExistingTextures 	= false,
	public ExportTexturesOnly			= false,
	public LoadAnimationOnly 			= false,
	
	-- Forward declarations
	public fn IsValidFile aFilePath = (),
	public fn SetFile aFilePath = (),
	public fn Load = (),	
	
	-- Animation
	private fn ReadAnimationGroup chunkDesc = #(),
	private fn ReadAnimation = #(),
	
	-- Texture 
	private fn ReadTextureList chunkDesc = (),
	private fn ReadFieldTexturePack chunkDesc = (),
	private fn ReadFieldTexture ddsOutPath texPakStream = (),
	
	-- Material
	private fn ReadMaterialList chunkDesc = (),
	private fn TestMaterialEndSkip endSkip isLast = (),
	
	-- Scene 
	private fn ReadScene chunkDesc = (),
	private fn ReadNode = (),
	private fn ReadMesh = (),
	private fn ReadParticle = (),
	
	-- Setup
	private fn SetupAnimation anim = (),
	private fn SetupNodeBone gfsNode = (),
	private fn SetupScene = (),
	
	public fn IsValidFile filePath =
	(
		local result = false
		local reader = FileStreamReader()
		reader.SetFile filePath true
		
		if (reader.GetFileStreamLength() > 4) then
		(
			local fileMagic = reader.ReadU32()
			reader.SeekCur -4
			if (fileMagic == GFSFileMagic) then
				result = true
		)
		
		reader.Close()
		
		result
	),
	
	public fn SetFile aFilePath =
	(
		m_FileInfo = FileUtils.GetFileInfo aFilePath
		if (m_FileInfo.FileExists) then
		(
			if (not (IsValidFile aFilePath)) then throw "GFSFileLoader.SetFile: File is not a valid GFS file."
			m_Reader = FileStreamReader()
			m_Reader.SetFile aFilePath true
		)
	),
	
	public fn Load =
	(
		m_Reader.SeekCur(4)
		m_Gfs = GFSFile()
		if ((m_FileInfo.FileExtension == ".GFS") and (DoesDirectoryExist (m_FileInfo.FileDirectory + "textures"))) then
		(
			m_LoadExternalTex = true
		)
		else if (m_FileInfo.FileExtension == ".GAP") then 
		(
			LoadAnimationOnly = true
		)
		
		do
		(
			local desc = GFSChunkDescriptor()
			desc.StartOffset = m_Reader.GetPosition()
			desc.Version = m_Reader.ReadU32()			
			desc.Type = m_Reader.ReadU32()
			desc.Size = m_Reader.ReadU32()
			Format "GFSFileLoader.Load: Chunk % ver: % sz: % at %\n" (HexStr(desc.Type)) (HexStr(desc.Version)) (HexStr(desc.Size)) (HexStr(desc.StartOffset))
			
			case (desc.Type) of
			(
				GFSChunkTypeFileStart:
				(
					Format "GFSFileLoader.Load: GFS start marker chunk\n"
				)
			
				GFSChunkTypeAnimationGroup:
				(
					if (LoadAnimationOnly) then
					(
						ReadAnimationGroup(desc)
						
						/* HACK: Force stop */
						desc.Type = GFSChunkTypeFileEnd
					)
					else
					(
						m_Reader.SeekCur (desc.Size - 12)
					)
				)
				
				GFSChunkTypeTextureList:
				(
					if (not LoadAnimationOnly) then
					(
						if (not m_LoadExternalTex) then 
						(
							ReadTextureList(desc)
						)
						else
						(
							ReadFieldTexturePack(desc)
							m_Reader.SeekCur (desc.Size - 12)
						)
					)
					else
					(
						m_Reader.SeekCur (desc.Size - 12)
					)
				)
				
				GFSChunkTypeMaterialList:
				(
					if ((not ExportTexturesOnly) and (not LoadAnimationOnly)) then
					(
						ReadMaterialList(desc)
					)
					else
					(
						m_Reader.SeekCur (desc.Size - 12)
					)
				)
				
				GFSChunkTypeScene:
				(
					if ((not ExportTexturesOnly) and (not LoadAnimationOnly)) then
					(
						ReadScene(desc)
					)
					else
					(
						m_Reader.SeekCur (desc.Size - 12)
					)
				)
				
				GFSChunkTypeFileEnd:
					Format "GFSFileLoader.Load: GFS end marker chunk\n"
				
				default:
					if (desc.Size != 0) then m_Reader.SeekCur (desc.Size - 12)
			)
		
			Format "\n"
		) while (desc.Type != GFSChunkTypeFileEnd)
		
		if ((not ExportTexturesOnly) and (not LoadAnimationOnly) and (m_Gfs.Scene != undefined)) then
		(
			SetupScene()
		)
		else if (LoadAnimationOnly and (m_Gfs.AnimationGroup != undefined)) then
		(
			SetupAnimation(m_Gfs.AnimationGroup.Animations[1])
		)
		
		m_Gfs
	),
	
	-- Animation
	private fn ReadAnimationGroup chunkDesc = 
	(
		m_Gfs.AnimationGroup = GFSAnimationGroup()
		m_Gfs.AnimationGroup.Field1 = m_Reader.ReadU32()
		m_Gfs.AnimationGroup.Flags = m_Reader.ReadU32()
		m_Gfs.AnimationGroup.AnimationCount = m_Reader.ReadU32()
		
		Format "\tGFSFileLoader.ReadAnimationGroup: f1: % flg: % animCount: %\n" m_Gfs.AnimationGroup.Field1 (HexStr(m_Gfs.AnimationGroup.Flags)) m_Gfs.AnimationGroup.AnimationCount
		
		m_Gfs.AnimationGroup.AnimationCount = 1
		for i = 1 to m_Gfs.AnimationGroup.AnimationCount do
		(
			local anim = ReadAnimation()
			m_Gfs.AnimationGroup.Animations[i] = anim
		)
		
		m_Reader.SeekCur(0x4) -- ?
	),
	
	private fn ReadAnimation = 
	(
		local anim = GFSAnimation()
		anim.Flags = m_Reader.ReadU32()
		anim.Duration = m_Reader.ReadF32()
		anim.ControllerCount = m_Reader.ReadU32()
		Format "\t\tGFSFileLoader.ReadAnimation: flg: % duration: % ControllerCount: %\n" (HexStr(anim.Flags)) anim.Duration anim.ControllerCount
		
		for i = 1 to anim.ControllerCount do
		(
			local controllerType = m_Reader.ReadU16()
			local ctrl = undefined
			
			case controllerType of 
			(
				1: -- PRS controller
				(
					ctrl = GFSAnimationPRSController()
					ctrl.NodeID = m_Reader.ReadU32()
					ctrl.NodeName = m_Reader.ReadU16PrefixedLengthString()
					ctrl.NodeHash = m_Reader.ReadU32()
					ctrl.Field2 = m_Reader.ReadU32()
					ctrl.KeyFrameFlags = m_Reader.ReadU32()
					ctrl.KeyFrameCount = m_Reader.ReadU32()
					Format "\t\t\tGFSFileLoader.ReadAnimation: PRSController (idx:%): % id: % f2: % keyFrameFlags: % keyFrameCount: %\n" \
									i ctrl.NodeName ctrl.NodeID (HexStr(ctrl.Field2)) (HexStr(ctrl.KeyFrameFlags)) ctrl.KeyFrameCount
					
					for j = 1 to ctrl.KeyFrameCount do
					(
						append ctrl.KeyFrameTimings (m_Reader.ReadF32())
					)
					
					for j = 1 to ctrl.KeyFrameCount do
					(
						local keyFrame = undefined
						case ctrl.KeyFrameFlags of
						(
							0x02:
							(
								m_Reader.SeekCur(0x10)
							)
							
							0x10:
							(
								m_Reader.SeekCur(0x4) -- float
							)
							
							0x1B:
							(
								keyFrame = GFSAnimationKeyFramePRS()
								keyFrame.Position = m_Reader.ReadVec3F16()
								keyFrame.Rotation = m_Reader.ReadQuatF16()
								keyFrame.Scale = m_Reader.ReadVec3F16()
							)
							
							0x1C:
							(
								keyFrame = GFSAnimationKeyFramePR()
								keyFrame.Position = m_Reader.ReadVec3F16()
								keyFrame.Rotation = m_Reader.ReadQuatF16()
							)
							
							default:
								throw ("GFSFileLoader.ReadAnimation: No handler for PRSControl keyframe type " + (ctrl.KeyFrameFlags as string) + " at offset " + (HexStr(m_Reader.GetPosition())))
						)
						
						append ctrl.KeyFrames keyFrame
					)
					
					if (ctrl.KeyFrameFlags == 2 or ctrl.KeyFrameFlags == 0x1B or ctrl.KeyFrameFlags == 0x1C) then
					(
						ctrl.BasePosition = m_Reader.ReadVec3F32()
						ctrl.BaseScale = m_Reader.ReadVec3F32()
					)
				)
				
				2: -- Material animation 
				(
					ctrl = GFSAnimationMaterialController()
					ctrl.MaterialID = m_Reader.ReadU32()
					ctrl.MaterialName = m_Reader.ReadU16PrefixedLengthString()
					ctrl.MaterialHash = m_Reader.ReadU32()
					ctrl.Field2 = m_Reader.ReadU32()
					ctrl.KeyFrameFlags = m_Reader.ReadU32()
					ctrl.KeyFrameCount = m_Reader.ReadU32()
					Format "\t\t\tGFSFileLoader.ReadAnimation: MaterialController (idx:%): % id: % f2: % keyFrameFlags: % keyFrameCount: %\n" \
									i ctrl.MaterialName ctrl.MaterialID (HexStr(ctrl.Field2)) (HexStr(ctrl.KeyFrameFlags)) ctrl.KeyFrameCount
					
					for j = 1 to ctrl.KeyFrameCount do
					(
						append ctrl.KeyFrameTimings (m_Reader.ReadF32())
					)
					
					for j = 1 to ctrl.KeyFrameCount do
					(
						local keyFrame = undefined
						case ctrl.KeyFrameFlags of
						(
							0xC:
							(
								m_Reader.SeekCur(0x4) -- float
							)
							
							0x1D:
							(
								m_Reader.SeekCur(0x4) -- float
							)
							
							default:
								throw ("GFSFileLoader.ReadAnimation: No handler for MaterialControl keyframe type " + (ctrl.KeyFrameFlags as string) + " at offset " + (HexStr(m_Reader.GetPosition())))
						)
						
						append ctrl.KeyFrames keyFrame
					)
					
					if (ctrl.Field2 == 2) then 
						m_Reader.SeekCur(0x18) -- more keyframes?
				)
				
				3: -- Camera animation 
				(
					ctrl = GFSAnimationCameraController()
					ctrl.CameraID = m_Reader.ReadU32()
					ctrl.CameraName = m_Reader.ReadU16PrefixedLengthString()
					ctrl.CameraHash = m_Reader.ReadU32()
					ctrl.Field2 = m_Reader.ReadU32()
					ctrl.KeyFrameFlags = m_Reader.ReadU32()
					ctrl.KeyFrameCount = m_Reader.ReadU32()
					Format "\t\t\tGFSFileLoader.ReadAnimation: CameraController (idx:%): % id: % f2: % keyFrameFlags: % keyFrameCount: %\n" \
									i ctrl.CameraName ctrl.CameraID (HexStr(ctrl.Field2)) (HexStr(ctrl.KeyFrameFlags)) ctrl.KeyFrameCount
					
					for j = 1 to ctrl.KeyFrameCount do
					(
						append ctrl.KeyFrameTimings (m_Reader.ReadF32())
					)
					
					for j = 1 to ctrl.KeyFrameCount do
					(
						local keyFrame = undefined
						case ctrl.KeyFrameFlags of
						(
							0x17:
							(
								m_Reader.SeekCur(0x4) #seek_cur -- float
							)
							
							default:
								throw ("GFSFileLoader.ReadAnimation: No handler for CameraControl keyframe type " + (ctrl.KeyFrameFlags as string) + " at offset " + (HexStr(m_Reader.GetPosition())))
						)
						
						append ctrl.KeyFrames keyFrame
					)
				)
				
				5: -- Morph
				(
					ctrl = GFSAnimationMorphController()
					ctrl.NodeID = m_Reader.ReadU32()
					ctrl.NodeName = m_Reader.ReadU16PrefixedLengthString()
					ctrl.NodeHash = m_Reader.ReadU32()
					ctrl.Field2 = m_Reader.ReadU32()
					ctrl.KeyFrameFlags = m_Reader.ReadU32()
					ctrl.KeyFrameCount = m_Reader.ReadU32()
					Format "\t\t\tGFSFileLoader.ReadAnimation: MorphController (idx:%): % id: % f2: % keyFrameFlags: % keyFrameCount: %\n" \
									i ctrl.NodeName ctrl.NodeID (HexStr(ctrl.Field2)) (HexStr(ctrl.KeyFrameFlags)) ctrl.KeyFrameCount
					
					for j = 1 to ctrl.KeyFrameCount do
					(
						append ctrl.KeyFrameTimings (m_Reader.ReadF32())
					)
					
					for j = 1 to ctrl.KeyFrameCount do
					(
						local keyFrame = undefined
						case ctrl.KeyFrameFlags of
						(
							0x5:
							(
								m_Reader.SeekCur(0x4) #seek_cur -- float
							)
							
							default:
								throw ("GFSFileLoader.ReadAnimation: No handler for MorphControl keyframe type " + (ctrl.KeyFrameFlags as string) + " at offset " + (HexStr(m_Reader.GetPosition())))
						)
						
						append ctrl.KeyFrames keyFrame
					)
				)
				
				default:
					throw ("GFSFileLoader.ReadAnimation: No handler for animation controller type " + (controllerType as string) + " at offset " + (HexStr(m_Reader.GetPosition())))
			)
			
			Format "\t\t\tGFSFileLoader.ReadAnimation: End at %\n\n" (HexStr(m_Reader.GetPosition()))
			
			if (ctrl != undefined) then
				ctrl.Type = controllerType
			
			anim.Controllers[i] = ctrl
		)
		
		m_Reader.SeekCur(0x18) 

		anim
	),
		
	-- Texture List 	
	private fn ReadTextureList chunkDesc =
	(
		m_Gfs.Textures = #()
		local field0 = m_Reader.ReadU32()
		if (field0 != 0) then
		(
			Format "\tGFSFileLoader.ReadTextureList: Unexpected value for field0: %\n" field0
		)
		
		local numTextures = m_Reader.ReadU32()
		Format "\tGFSFileLoader.ReadTextureList: Loading % textures\n\n" numTextures
		
		for i = 1 to numTextures do
		(
			local texture = GFSTexture()
			texture.Name = m_Reader.ReadU16PrefixedLengthString()
			texture.PixelFormat = m_Reader.ReadU16()
			texture.DataLength = m_Reader.ReadU32()
			Format "\tGFSFileLoader.ReadTextureList: Texture \"%\" fmt % sz % off %\n" texture.Name (HexStr(texture.PixelFormat)) (HexStr(texture.DataLength)) (HexStr(m_Reader.GetPosition()))
			
			local textureDDSPath = (m_FileInfo.FileDirectory + texture.Name)
				
			if (OverwriteExistingTextures or (not (DoesFileExist textureDDSPath))) then
			(
				Format "\tGFSFileLoader.ReadTextureList: Writing to %\n\n" textureDDSPath
				FileUtils.FileStreamBlockCopyToFile (m_Reader.GetFileStream()) textureDDSPath texture.DataLength
			)
			else
			(
				Format "\tGFSFileLoader.ReadTextureList: % already exists, skipping\n\n" textureDDSPath
				m_Reader.SeekCur(texture.DataLength)
			)
			
			local footer = m_Reader.ReadU32()
			if (footer != GFSTextureFooterMagic) then
			(
				throw "\tGFSFileLoader.ReadTextureListChunk: Unexpected end of texture payload data"
			)
			
			append m_Gfs.Textures texture
		)
	),
	
	private fn ReadFieldTexturePack chunkDesc =
	(
		-- Create texture pack file name
		local texturePackName = replace (m_FileInfo.FileName + "_00.bin") 1 1 "tex"
		texturePackName = replace (texturePackName) 12 0 "0"
		
		-- Create texture pack full path
		local texturePackPath = m_FileInfo.FileDirectory + "textures\\" + texturePackName
		local texPakFInfo = FileUtils.GetFileInfo texturePackPath
		local texPakReader = FileStreamReader()
		texPakReader.SetFile texturePackPath false
		
		Format "\tGFSFileLoader.ReadFieldTexturePack: Reading field texture pack %\n" texturePackPath
		
		while (true) do
		(
			texPakReader.SetBigEndian(false)
			local pakEntryName = texPakReader.ReadFixedLengthString 252
			local pakEntryDataSize = texPakReader.ReadU32()
			local pakEntryDataEnd = texPakReader.GetPosition() + pakEntryDataSize
			pakEntryDataEnd = bit.and (pakEntryDataEnd + 63) (bit.not 63) -- make sure to align the offset
			
			Format "\n\tGFSFileLoader.ReadFieldTexturePack: Pack entry: % size:%\n" pakEntryName (HexStr(pakEntryDataSize))
			
			if (pakEntryDataSize == 0) then
			(
				Format "\tGFSFileLoader.ReadFieldTexturePack: End of file marker read\n"
				exit
			)
			
			if ((GetFilenameType pakEntryName) != ".dds") then
			(
				Format "\tGFSFileLoader.ReadFieldTexturePack: Not a dds file, skipping.\n"
			)
			else
			(
				local ddsPath = (texPakFInfo.FileDirectory + pakEntryName)
				if (OverwriteExistingTextures or (not (DoesFileExist ddsPath))) then
				(
					ReadFieldTexture ddsPath texPakReader
				)
				else
				(
					Format "\tGFSFileLoader.ReadFieldTexturePack: % already exists, skipping\n\n" ddsPath
				)
			)
			
			texPakReader.SeekSet(pakEntryDataEnd)
		)
	),
	
	private fn ReadFieldTexture ddsOutPath texPakReader =
	(
		-- Read field texture header
		texPakReader.SetBigEndian(true)
		local headerStart = texPakReader.GetPosition()
		local magic = texPakReader.ReadU32()
		local dataSize = texPakReader.ReadU32()
		local field3 = texPakReader.ReadU32()
		local field4 = texPakReader.ReadU32()
		local dataStart = texPakReader.ReadU32()
		local dataSize2 = texPakReader.ReadU32()
		local field7 = texPakReader.ReadU32()
		local field8 = texPakReader.ReadU32()
		local width = texPakReader.ReadU16()
		local height = texPakReader.ReadU16()
		local field11 = texPakReader.ReadU16()
		
		texPakReader.SeekSet(headerStart + dataStart)
		
		Format "\tGFSFileLoader.ReadFieldTexture: Converting texture sz: % f3: % f4: % sz2: % f7: % f8: % w: % h: % f11: %\n\n" \
					(HexStr(dataSize)) (HexStr(field3)) (HexStr(field4)) (HexStr(dataSize2)) (HexStr(field7)) \
					(HexStr(field8)) width height (HexStr(field11))
			
		-- Create dds header						
		local ddsHdr = DDSHeader()
		ddsHdr.Width = width
		ddsHdr.Height = height
		ddsHdr.MipMapCount = (bit.shift (bit.and field7 0xFF0000) -16)
						
		if ((bit.and field7 0x1000000) == 0x1000000) then
		(
			ddsHdr.PixelFormatFourCC = DDSPixelFormatFourCCDXT3
			bit.or ddsHdr.PixelFormatFlags DDSPixelFormatFlagAlphaPixels
		)
		else if ((bit.and field7 0x8000000) == 0x8000000) then -- Could also be a cubemap
		(
			ddsHdr.PixelFormatFourCC = DDSPixelFormatFourCCDXT5
			bit.or ddsHdr.PixelFormatFlags DDSPixelFormatFlagAlphaPixels
		)
		
		-- Write dds data
		local ddsStream = FileUtils.CreateWriteableFileStream(ddsOutPath)
		ddsHdr.WriteToStream(ddsStream)
		FileUtils.FileStreamBlockCopy (texPakReader.GetFileStream()) ddsStream dataSize
	),
	
	-- Material List		
	private fn ReadMaterialList chunkDesc = 
	(
		try
		(
			
		m_Gfs.Materials = #()
		local field0 = m_Reader.ReadU32()
		if (field0 != 0) then
		(
			Format "\tGFSFileLoader.ReadMaterialList: Unexpected value for field0: %\n" field0
		)
	
		local numMaterials = m_Reader.ReadU32()
		Format "\tGFSFileLoader.ReadMaterialList: Loading % materials\n" numMaterials
		
		for i = 1 to numMaterials do 
		(
			material = GFSMaterial()
			material.Name = m_Reader.ReadU16PrefixedLengthString()
			material.Hash = m_Reader.ReadU32()
			material.Flags = m_Reader.ReadU32()
			material.FloatFields = m_Reader.ReadF32Array 18
			material.ShortFields = m_Reader.ReadU16Array 8
			material.UintFields = m_Reader.ReadU32Array 2
			material.ShortField = m_Reader.ReadU16()
			material.UintField = m_Reader.ReadU32()
			Format "\n\tGFSFileLoader.ReadMaterialList: Material % hash: % flgs: % off %\n" material.Name (HexStr(material.Hash)) (HexStr(material.Flags)) (HexStr(m_Reader.GetPosition()))

			for j = 1 to GFSMaterialTextureSlotCount do
			(
				local textureSlot = undefined
				local bitMask = bit.shift GFSMaterialFlagsUseDiffuseTextureSlot (j - 1)
				if ((bit.and material.Flags bitMask) == bitMask) then
				(
					textureSlot = GFSMaterialTextureSlot()
					textureSlot.TextureName = m_Reader.ReadU16PrefixedLengthString()
					textureSlot.Hash = m_Reader.ReadU32()
					m_Reader.SeekCur(0x48)
					
					Format "\t\tGFSFileLoader.ReadMaterialList: % \"%\"\n" (GFSMaterialFlagsTextureSlotNames[j]) textureSlot.TextureName
				)
				
				material.TextureSlots[j] = textureSlot
			)
			
			local isLast = (i == numMaterials)
			local foundEndSkip = TestMaterialEndSkip 0x00 isLast
			if (not foundEndSkip) then foundEndSkip = TestMaterialEndSkip 0x10 isLast
			if (not foundEndSkip) then foundEndSkip = TestMaterialEndSkip 0x3C isLast
			if (not foundEndSkip) then foundEndSkip = TestMaterialEndSkip 0x48 isLast
			if (not foundEndSkip) then 
			(
				for j = 1 to 0x100 do
				(
					if ((j == 0x10) OR (j == 0x3C) OR (j == 0x48)) then continue
					
					foundEndSkip = TestMaterialEndSkip j isLast
					if (foundEndSkip) then
						exit
				)
			)
			if (not foundEndSkip) then
			(
				Format "\t\tGFSFileLoader.ReadMaterialList: Couldn't find material end skip"
			)
			
			append m_Gfs.Materials material
		)

		) -- end try
		catch
		(
			Format "\tGFSFileLoader.ReadMaterialList: Failed to load materials\n"
			m_Gfs.Materials = undefined
			m_Reader.SeekSet (chunkDesc.StartOffset + chunkDesc.Size)
		)
	),
	
	private fn TestMaterialEndSkip endSkip isLast =
	(
		local startPos = m_Reader.GetPosition()
		
		try
		(
			m_Reader.SeekCur(endSkip)
			
			if (not isLast) then
			(
				local nameLen = m_Reader.ReadU16()
				if (nameLen == 0 OR nameLen > 64) then
				(
					m_Reader.SeekSet(startPos)
					return false
				)
				
				for i = 1 to nameLen do
				(
					local curByte = m_Reader.ReadU8()
					if ((curByte == 0) OR (curByte > 0x7F)) then
					(
						m_Reader.SeekSet(startPos)
						return false
					)
				)
				
				m_Reader.SeekCur(0x58)
				if ((m_Reader.ReadU16()) != 4) then
				(
					m_Reader.SeekSet(startPos)
					return false
				)
			)
			else
			(
				local ver = m_Reader.ReadU32()
				if (not ((ver >= 0x1100000) AND (ver < 0x11FFFFF))) then
				(
					m_Reader.SeekSet(startPos)
					return false
				)
			)
		)
		catch
		(
			m_Reader.SeekSet(startPos)
			return false
		)
	
		m_Reader.SeekSet(startPos + endSkip)
		return true
	),
	
	-- Scene
	private fn ReadScene chunkDesc =
	(
		m_Gfs.Scene = GFSScene()
		m_Gfs.Scene.Field0 = m_Reader.ReadU32()
		m_Gfs.Scene.Flags = m_Reader.ReadU32()
		
		Format "\tGFSFileLoader.ReadScene: Scene field0: % flags: %\n" m_Gfs.Scene.Field0 m_Gfs.Scene.Flags
		
		if ((bit.and m_Gfs.Scene.Flags GFSSceneFlagsSkinned) == GFSSceneFlagsSkinned) then
		(
			m_Gfs.Scene.UsedBoneCount = m_Reader.ReadU32()
			Format "\tGFSFileLoader.ReadScene: Skinning is used, reading data for % skinned nodes\n" m_Gfs.Scene.UsedBoneCount
			
			m_Gfs.Scene.InvBindMatrices = m_Reader.ReadMtx44Array m_Gfs.Scene.UsedBoneCount
			m_Gfs.Scene.UsedBoneMap 	= m_Reader.ReadU16Array   m_Gfs.Scene.UsedBoneCount
		)
		
		if ((bit.and m_Gfs.Scene.Flags GFSSceneFlagsHasBBox) == GFSSceneFlagsHasBBox) then
		(
			Format "\tGFSFileLoader.ReadScene: Bbox data present. Skipping.\n"
			m_Reader.SeekCur(0x28) -- Bbox data
		)
		
		-- Create root node
		local gfsRootNode = ReadNode()
			
		local parentStack = StackContainer()
		parentStack.PushItem(gfsRootNode)
			
		local remChildCountStack = StackContainer()
		remChildCountStack.PushItem(gfsRootNode.ChildCount)
		
		do
		(
			local node = ReadNode()
			node.Parent = parentStack.PeekItem()
			
			-- Check remaining children to read of parent
			local remChildCount = remChildCountStack.PopItem()
			remChildCount -= 1
			
			if (remChildCount == 0) then
			(
				parentStack.PopItem()
			)
			else
			(
				remChildCountStack.PushItem(remChildCount)
			)

			/* Add to parent's children list */
			append node.Parent.Children node
			
			if (node.ChildCount > 0) then
			(
				parentStack.PushItem(node)
				remChildCountStack.PushItem(node.ChildCount)
			)
			
		) while (not remChildCountStack.IsEmpty())
			
		local curNodeStack = StackContainer()
		curNodeStack.PushItem(gfsRootNode)
		
		while (not curNodeStack.IsEmpty()) do
		(
			local curNode = curNodeStack.PopItem()			
			append m_Gfs.Scene.Nodes curNode
			
			for i = 1 to curNode.ChildCount do
			(
				curNodeStack.PushItem (curNode.Children[i])
			)
		)
	),
			
	private fn ReadNode =
	(
		Format "\n\t\tGFSFileLoader.ReadNode: Reading node at %\n" (HexStr(m_Reader.GetPosition()))
		node = GFSNode()
		node.Name = m_Reader.ReadU16PrefixedLengthString()
		node.Hash = m_Reader.ReadU32()
		node.Position = m_Reader.ReadVec3F32()
		node.Rotation = m_Reader.ReadQuatF32()
		node.Scale = m_Reader.ReadVec3F32()
		node.AttachmentCount = m_Reader.ReadU32()

		Format "\t\tGFSFileLoader.ReadNode: Node \"%\" hash: % attachmentCount: % \n" node.Name (HexStr(node.Hash)) node.AttachmentCount		
		for i = 1 to node.AttachmentCount do
		(
			local attachment = GFSNodeAttachment()
			attachment.Type = m_Reader.ReadU32()
			
			Format "\n\t\tGFSFileLoader.ReadNode: Attachment #% Type: % \n" i attachment.Type
			case (attachment.Type) of
			(
				4: -- Mesh
				(
					attachment.Object = ReadMesh()
				)
					
				5: -- Camera
				(
					m_Reader.SeekCur(0x54)  -- Matrix4x4 + 5 floats?
				)
					
				6: -- Sky
				(
					local field1 = m_Reader.ReadU32()
					local field2 = m_Reader.ReadU32()
					if (field1 == 3 AND field2 == 2) then m_Reader.SeekCur 0x44
					else if (field1 == 3 AND field2 == 3) then m_Reader.SeekCur 0x58
					else if (field1 == 1 AND field2 == 2) then m_Reader.SeekCur 0x48
					else if (field1 == 1 AND field2 == 3) then m_Reader.SeekCur 0x5C
					else throw ("GFSFileLoader.ReadNode: Can't handle this attachment type f1: " + (field1 as string) + " f2: " + field2 as string)
					--FSeek m_Stream 0x50 #seek_cur
				)
				
				7: -- Epl particle
				(
					local offset = m_Reader.GetPosition()
					print (HexStr(offset))
						
					if ((offset == 0x1a67d5) OR (offset == 0x1b09ef)) then
					(
						m_Reader.SeekCur(0x27C9)
					)
					else if ((offset == 0x1a907c) OR (offset == 0x1b3296)) then
					(
						m_Reader.SeekCur(0x3568)
					)
					else if ((offset == 0x1ac6c2) OR (offset == 0x1b68dc)) then
					(
						m_Reader.SeekCur(0x3571)
					)
					else
					(
						throw "\t\tGFSFileLoader.ReadNode: Can't handle this attachment type\n"
					)
					--ReadParticle()
				)
				
				/*
				9: -- Morpher
				(
					
				)
				*/

				default:
				(
					throw "\t\tGFSFileLoader.ReadNode: Unknown attachment type\n"
				)
			)
			
			node.Attachments[i] = attachment
		)
		
		/* Read user properties */
		local hasProps = m_Reader.ReadU8() == 1
		if (hasProps) then
		(
			local numProps = m_Reader.ReadU32()
			Format "\n\t\tGFSFileLoader.ReadNode: Node has % user properties.\n" numProps

			for i = 1 to numProps do
			(
				local prop = GFSNodeUserProperty()
				prop.Type = m_Reader.ReadU32()
				prop.Key = m_Reader.ReadU16PrefixedLengthString()
				prop.Hash = m_Reader.ReadU32()
				prop.Size = m_Reader.ReadU32()
				
				case (prop.Type) of
				(
					GFSNodeUserPropertyTypeInt:
						prop.Value = m_Reader.ReadU32()
					
					GFSNodeUserPropertyTypeFloat:
						prop.Value = m_Reader.ReadF32()
					
					GFSNodeUserPropertyTypeBool:
					(
						prop.Value = m_Reader.ReadU8()
						if (prop.Value != 0 and prop.Value != 1) then
						(
							Format "\t\tGFSFileLoader.ReadNode: GFSNodeUserPropertyTypeBool value out of boolean range. Value %\n" prop.Value
						)
						else
						(
							prop.Value = (prop.Value == 1)
						)
					)
					
					GFSNodeUserPropertyTypeString:
						prop.Value = m_Reader.ReadFixedLengthString(prop.Size - 1)
					
					default:
					(
						Format "\t\tGFSFileLoader.ReadNode: Unknown user property type %\n" prop.Type
						m_Reader.SeekCur attrib.Size
					)
				)
				
				append node.UserProperties prop
				
				Format "\t\tGFSFileLoader.ReadNode: User property %  type: % hash: % size: % value: %\n" prop.Key prop.Type (HexStr(prop.Hash)) prop.Size prop.Value
			)
		)
		
		local lastFloatVal = (m_Reader.ReadF32())
		if (lastFloatVal != 1.0f) then
			Format "\t\tGFSFileLoader.ReadNode: Last float value isn't 1.0f: %\n" lastFloatVal
		
		node.ChildCount = m_Reader.ReadU32()

		node
	),
	
	private fn ReadMesh = 
	(
		mesh = GFSMesh()
		mesh.Flags = m_Reader.ReadU32()
		mesh.VertexFormat = m_Reader.ReadU32()
		mesh.FaceCount = m_Reader.ReadU32()
		mesh.Field5 = m_Reader.ReadU16()
		mesh.VertexCount = m_Reader.ReadU32()
		mesh.Field7 = m_Reader.ReadU32()
		mesh.VertexTexCoordChannels = #(#(), #(), #(), #(), #(), #(), #(), #())

		Format "\n\t\t\tGFSFileLoader.ReadMesh: F % FVF % FC % VC % [FVF Flags] NRM % TAN % CLR % TEX1 % TEX2 % TEX3 % TEX4 % TEX5 % TEX6 % TEX7 % TEX8 % WGHT %\n" \
			(HexStr(mesh.Flags)) (HexStr(mesh.VertexFormat)) mesh.FaceCount mesh.VertexCount \
			((bit.and mesh.VertexFormat GFSMeshFVFNormal) == GFSMeshFVFNormal) \
			((bit.and mesh.VertexFormat GFSMeshFVFTangent) == GFSMeshFVFTangent) \
			((bit.and mesh.VertexFormat GFSMeshFVFColor) == GFSMeshFVFColor) \
			((bit.and mesh.VertexFormat GFSMeshFVFTex1) == GFSMeshFVFTex1) \
			((bit.and mesh.VertexFormat GFSMeshFVFTex2) == GFSMeshFVFTex2) \
			((bit.and mesh.VertexFormat GFSMeshFVFTex3) == GFSMeshFVFTex3) \
			((bit.and mesh.VertexFormat GFSMeshFVFTex4) == GFSMeshFVFTex4) \
			((bit.and mesh.VertexFormat GFSMeshFVFTex5) == GFSMeshFVFTex5) \
			((bit.and mesh.VertexFormat GFSMeshFVFTex6) == GFSMeshFVFTex6) \
			((bit.and mesh.VertexFormat GFSMeshFVFTex7) == GFSMeshFVFTex7) \
			((bit.and mesh.VertexFormat GFSMeshFVFTex8) == GFSMeshFVFTex8) \
			((bit.and mesh.Flags 1) == 1) 
			
		Format "\t\t\tGFSFileLoader.ReadMesh: Vertex start %\n" (HexStr(m_Reader.GetPosition()))			
		for i = 1 to mesh.VertexCount do
		(
			mesh.VertexPositions[i] = m_Reader.ReadVec3F32()
			
			if ((bit.and mesh.VertexFormat GFSMeshFVFNormal) == GFSMeshFVFNormal) then
				mesh.VertexNormals[i] = m_Reader.ReadVec3F32()
			
			if ((bit.and mesh.VertexFormat GFSMeshFVFTangent) == GFSMeshFVFTangent) then
				mesh.VertexTangents[i] = m_Reader.ReadVec3F32()
			
			if ((bit.and mesh.VertexFormat GFSMeshFVFColor) == GFSMeshFVFColor) then
				mesh.VertexColors[i] = m_Reader.ReadU32()
			
			for j = 1 to 8 do
			(
				local bitMask = bit.shift GFSMeshFVFTex1 (j - 1)
				if ((bit.and mesh.VertexFormat bitMask) == bitMask) then
				(
					local texCoord = m_Reader.ReadVec2F32()
					mesh.VertexTexCoordChannels[j][i] = [texCoord.X, (texCoord.Y * -1) + 1, 0]
				)
			)

			if ((bit.and mesh.Flags 1) == 1) then
			(
				local weightTotal = 0.0f
				local numUsed = 0
				mesh.VertexWeights[i] = #()
				for j = 1 to 4 do
				(
					local weight = m_Reader.ReadF32()
					mesh.VertexWeights[i][j] = weight
					weightTotal += weight
					
					if (weight > 0.0f) then
						numUsed += 1
				)
				
				if (weightTotal < 0.998) then
				(
					mesh.VertexWeights[i][numUsed] += (1.0f - weightTotal)
				)
				
				mesh.VertexWeightIndices[i] = #()
				for j = 1 to 4 do
					mesh.VertexWeightIndices[i][4 - (j - 1)] = (m_Reader.ReadU8()) + 1
			)
		)
		Format "\t\t\tGFSFileLoader.ReadMesh: Faces start %\n" (HexStr(m_Reader.GetPosition()))
		mesh.FaceIndices = m_Reader.ReadVec3U16Array mesh.FaceCount
		for i = 1 to mesh.FaceCount do
			mesh.FaceIndices[i] = mesh.FaceIndices[i] + 1
		
		mesh.MaterialName = m_Reader.ReadU16PrefixedLengthString()
		mesh.MaterialHash = m_Reader.ReadU32()
		m_Reader.SeekCur(0x28)
		
		Format "\t\t\tGFSFileLoader.ReadMesh: Used material: \"%\" hash: %\n" mesh.MaterialName (HexStr(mesh.MaterialHash))
		Format "\t\t\tGFSFileLoader.ReadMesh: Mesh end at %\n" (HexStr(m_Reader.GetPosition()))
		mesh
	),
	
	private fn ReadParticle =
	(
		local numEffect = m_Reader.ReadU32()
		m_Reader.SeekCur(0x2F)
		
		for i = 1 to numEffect do
		(
			m_Reader.SeekCur(0x3E)
			local effectName = m_Reader.ReadU16PrefixedLengthString()
			print effectName
			m_Reader.SeekCur(0x14C)
			local textureName = m_Reader.ReadU16PrefixedLengthString()
			print textureName
			local textureHash = m_Reader.ReadU32()
			local unk1 = m_Reader.ReadU32()
			local unk2 = m_Reader.ReadU32()
			local textureDataSize = m_Reader.ReadU32()
			FileCopyBlockToFile m_Stream (m_FileInfo.FileDirectory + textureName) textureDataSize
			
			Print (HexStr(m_Reader.GetPosition()))
			
			m_Reader.SeekCur(1)
		)
		
		-- 0xCC = 5 particles
		m_Reader.SeekCur(0xCC)
	),
	
	-- Setup
	private fn SetupAnimation anim =
	(
		local numAnimFrames = (anim.Duration * 30) as integer
		animationRange = interval 0 numAnimFrames
		Format "\tGFSFileLoader.SetupAnimation: Setting up animation with % frames\n" numAnimFrames
			
		with animate on
		(
			for ctrl in anim.Controllers do
			(
				if ((ctrl == undefined) or (ctrl.Type != 1)) then
					continue -- Skip if not PRS
				
				local curBoneNode = GetNodeByName ctrl.NodeName exact: true
				if (curBoneNode == undefined) then 
					continue
				
				Format "\tGFSFileLoader.SetupAnimation: Setting up keyframes for % keyFrameCount: %\n" ctrl.NodeName ctrl.KeyFrameCount
				
				--local initPos = curBoneNode.Controller.Position
				local initRot = curBoneNode.Controller.Rotation
				--local initScale = curBoneNode.Controller.Scale
				
				/*
				if (ctrl.BasePosition != undefined) then
					at time 0 in coordsys parent curBoneNode.Position = ctrl.BasePosition
				*/
				
				for keyIdx = 1 to ctrl.KeyFrameCount do
				(
					local curTime = ctrl.KeyFrameTimings[keyIdx] * 30
					local curKey = ctrl.KeyFrames[keyIdx]
					
					if (curKey == undefined) then 
						continue
					
					at time curTime
					(
						curBoneNode.Controller.Rotation = inverse curKey.Rotation
						curBoneNode.Controller.Position = ctrl.BasePosition * curKey.Position
						if (ctrl.KeyFrameFlags == 0x1B) then
							curBoneNode.Controller.Scale = ctrl.BaseScale * curKey.Scale
					)
				)
			)
		)
	),
	
	private fn SetupNodeBone gfsNode = 
	(
		/* Create max bone node for GFSNode */
		
		--local tfm = ((scalematrix gfsNode.Scale) *  (gfsNode.Rotation as matrix3)) * (transmatrix gfsNode.Position) 
	
		local tfm = (Inverse gfsNode.Rotation) as matrix3
		tfm *= ScaleMatrix gfsNode.Scale
		tfm.row4 = gfsNode.Position
			
		local newBone = BoneSys.CreateBone tfm.row4 (tfm.row4 + 0.01 * (normalize tfm.row1)) (normalize tfm.row3)
		newBone.Name = gfsNode.Name
		newBone.Width  = 0.01
		newBone.Height = 0.01
		newBone.Transform = tfm
		newBone.SetBoneEnable false 0
		newBone.Wirecolor = red
		newbone.Showlinks = true
		newBone.Pos.Controller      = TCB_position()
		newBone.Rotation.Controller = TCB_rotation()
						  
		if gfsNode.Parent != undefined do 
		(
			newBone.Parent = gfsNode.Parent.MaxBoneNode
			newBone.Transform *= gfsNode.Parent.MaxBoneNode.transform
		)
		
		for prop in gfsNode.UserProperties do
		(
			SetUserProp newBone prop.Key (prop.Value as string)
		)
		
		newBone
	),
	
	private fn SetupScene =
	(
		for gfsNode in m_Gfs.Scene.Nodes do
		(
			gfsNode.MaxBoneNode = SetupNodeBone(gfsNode)
		)
		
		for gfsNode in m_Gfs.Scene.Nodes do
		(
			for attachmentIdx = 1 to gfsNode.Attachments.Count do
			(
				if (gfsNode.Attachments[attachmentIdx].Type != 4) then continue
				
				local gfsMesh = gfsNode.Attachments[attachmentIdx].Object
				
				/* Create max mesh node */
				local maxMeshNode = Mesh Vertices:gfsMesh.VertexPositions Faces:gfsMesh.FaceIndices Name:(gfsNode.Name + "_Mesh" + attachmentIdx as string)
				maxMeshNode.Parent = gfsNode.MaxBoneNode
				
				Format "\tGFSFileLoader.SetupScene: Setting up mesh for %\n" maxMeshNode.Name

				--if (not ((bit.and gfsMesh.Flags2 0x80000000) == 0x80000000)) then
				(
					/* Transform mesh to model space */
					--CenterPivot maxMeshNode
					maxMeshNode.Transform *= gfsNode.MaxBoneNode.Transform
				)
				
				/* Set up material */
				maxMeshNode.Material = StandardMaterial()
				maxMeshNode.Material.ShowInViewport = true
				maxMeshNode.Material.Name = gfsMesh.MaterialName
				
				/* Apply vertex normals */
				if (ApplyNormals) then
				(
					for j = 1 to gfsMesh.VertexNormals.Count do 
					(
						SetNormal maxMeshNode j gfsMesh.VertexNormals[j]
					)
				)
				
				/* Apply up to 8 texture coordinate channels */
				MeshOp.SetNumMaps maxMeshNode 8
				for i = 1 to 8 do
				(
					MeshOp.DefaultMapFaces maxMeshNode i
					for j = 1 to gfsMesh.VertexTexCoordChannels[i].Count do
					(
						MeshOp.SetMapVert maxMeshNode i j gfsMesh.VertexTexCoordChannels[i][j]
					)
				)
				
				/* Apply vertex colors */
				if (ApplyColors and ((bit.and gfsMesh.VertexFormat GFSMeshFVFColor) == GFSMeshFVFColor)) then
				(
					setNumCPVVerts maxMeshNode gfsMesh.VertexCount
					setCVertMode maxMeshNode true
					setShadeCVerts maxMeshNode true
					defaultVCFaces maxMeshNode
					Meshop.SupportVAlphas maxMeshNode
					
					for j = 1 to gfsMesh.VertexColors.Count do
					(
						local clr = gfsMesh.VertexColors[j]
						local b = bit.and clr 0xFF
						local g = bit.shift (bit.and clr 0xFF00) -8
						local r = bit.shift (bit.and clr 0xFF0000) -16
						local a = bit.shift (bit.and clr 0xFF000000) -24
						SetVertColor maxMeshNode j [r, g, b]
						Meshop.SetVAlpha maxMeshNode j (a as float / 255f)
					)
				)

				/* Apply material textures */
				if (m_Gfs.Materials != undefined) then
				(
					local gfsMaterial = undefined
					for mat in m_Gfs.Materials do
					(
						if (mat.Hash == gfsMesh.MaterialHash) then
						(
							gfsMaterial = mat
						)
					)
				
					if (gfsMaterial != undefined) then
					(	
						maxMeshNode.Material.AdLock 	= off
						maxMeshNode.Material.Ambient 	= [gfsMaterial.FloatFields[1] * 255, gfsMaterial.FloatFields[2] * 255, gfsMaterial.FloatFields[3] * 255]
						maxMeshNode.Material.Diffuse 	= [gfsMaterial.FloatFields[5] * 255, gfsMaterial.FloatFields[6] * 255, gfsMaterial.FloatFields[7] * 255]						
						maxMeshNode.Material.Specular 	= [gfsMaterial.FloatFields[9] * 255, gfsMaterial.FloatFields[10] * 255, gfsMaterial.FloatFields[11] * 255]						
						--maxMeshNode.Material.UseSelfIllumColor = true
						maxMeshNode.Material.SelfIllumColor = [gfsMaterial.FloatFields[13] * 255, gfsMaterial.FloatFields[14] * 255, gfsMaterial.FloatFields[15]  * 255]
						
						local textureDirectory = m_FileInfo.FileDirectory
						if (m_LoadExternalTex) then
							textureDirectory += "textures\\"
									
						if (gfsMaterial.TextureSlots[GFSMaterialDiffuseTextureSlotID] != undefined) then
						(
							maxMeshNode.Material.DiffuseMap = BitmapTexture Filename:(textureDirectory + gfsMaterial.TextureSlots[GFSMaterialDiffuseTextureSlotID].TextureName) AlphaSource:0 MonoOutput:1
						)
						
						if (gfsMaterial.TextureSlots[GFSMaterialNormalTextureSlotID] != undefined) then
						(
							maxMeshNode.Material.BumpMapAmount = 100
							maxMeshNode.Material.BumpMap = Normal_Bump()
							maxMeshNode.Material.BumpMap.Normal_map = BitmapTexture Filename:(textureDirectory + gfsMaterial.TextureSlots[GFSMaterialNormalTextureSlotID].TextureName) AlphaSource:0 MonoOutput:1
						)
						
						if (gfsMaterial.TextureSlots[GFSMaterialSpecularTextureSlotID] != undefined) then
						(
							maxMeshNode.Material.SpecularLevelMapAmount = 25
							maxMeshNode.Material.SpecularLevelMap = BitmapTexture Filename:(textureDirectory + gfsMaterial.TextureSlots[GFSMaterialSpecularTextureSlotID].TextureName) AlphaSource:0 MonoOutput:1
						)
						
						if (gfsMaterial.TextureSlots[GFSMaterialReflectionTextureSlotID] != undefined) then
						(
							maxMeshNode.Material.ReflectionMap = BitmapTexture Filename:(textureDirectory + gfsMaterial.TextureSlots[GFSMaterialReflectionTextureSlotID].TextureName) AlphaSource:0 MonoOutput:1
						)
						
						if (gfsMaterial.TextureSlots[GFSMaterialHighlightTextureSlotID] != undefined) then
						(
							maxMeshNode.Material.SelfIllumMap = BitmapTexture Filename:(textureDirectory + gfsMaterial.TextureSlots[GFSMaterialHighlightTextureSlotID].TextureName) AlphaSource:0 MonoOutput:1
						)
					)
				)
				
				/* Apply skin */
				if (ApplySkin) then
				(						
					if (((Bit.And gfsMesh.Flags 1) == 1)) then
					(
						Format "\tGFSFileLoader.SetupScene: Setting up skin modifier\n"
						
						Max modify mode
						Select maxMeshNode
						skinMod = Skin()
						AddModifier maxMeshNode skinMod
						
						/* Get used skinning nodes */
						for i = 1 to m_Gfs.Scene.UsedBoneCount do
						(
							SkinOps.AddBone skinMod m_Gfs.Scene.Nodes[m_Gfs.Scene.UsedBoneMap[i] + 1].MaxBoneNode 0
						)
						
						/* Set the weight for each of them */
						ModPanel.SetCurrentObject skinMod
						for i = 1 to gfsMesh.VertexCount do 
						(
							SkinOps.SetVertexWeights skinMod i gfsMesh.VertexWeightIndices[i] gfsMesh.VertexWeights[i]
						)
						
						Max create mode
					)
				)
				
				Format "\n"
			)
		)
		
		if (ApplyYToZUp) then
		(
			Rotate m_Gfs.Scene.Nodes[1].MaxBoneNode (AngleAxis 90 [1,0,0])
		)
	)
)
